from mathreader.api import *
from mathreader.config import Configuration
from mathreader.helpers.exceptions import *
import base64
import numpy as np
import cv2
import sys

configs = Configuration()


class Example:

    def __init__(self):

        expression = ""
        hme_recognizer = HME_Recognizer()
        print(sys.argv)
        if sys.argv and len(sys.argv) > 1:
            images = [configs.package_path + '/images/' + sys.argv[1]]
        else:
            images = [configs.package_path + '/images/validation/21.png']

        for image in images:

            try:
                hme_recognizer.load_image(image, data_type='path')
                expression, img = hme_recognizer.recognize()

                lex_errors = hme_recognizer.get_lex_errors()
                yacc_errors = hme_recognizer.get_yacc_errors()
                pure_lex_errors = hme_recognizer.get_lex_pure_errors()
                pure_yacc_errors = hme_recognizer.get_yacc_pure_errors()
                latex_string_original = hme_recognizer.get_latex_string_original()

                print('\n\nLex errors: ', lex_errors)
                print('\n\nYacc errors: ', yacc_errors)
                print('\n\nPure Lex Errors:', pure_lex_errors)
                print('\n\nPure Yacc Errors: ', pure_yacc_errors)
                print('\n\nOriginal Expression: ', latex_string_original)

            except (GrammarError, SintaticError, LexicalError) as e:

                if 'latex_string_original' in e.data:
                    expression = e.data['latex_string_original']

                print('[example.py] Exception: ', e.data)
                print('[example.py] Exception: ', e.valor)

            print("\nExpression: ", expression)


Example()



from mathreader.image_processing import preprocessing as preprocessing
from mathreader import helpers
import numpy as np
import cv2
import imutils
import os


class ImagePostprocessing:

    def __init__(self, image):
        self.image = image

    def segment_equality(self, symbols):
        xmin_sorted = sorted(symbols, key=lambda i: i['xmin'], reverse=False)
        xmin_ymin_sorted = sorted(symbols, key=lambda i: (i['xmin'],
                                  i['ymin']), reverse=False)

        try:
            c = 0
            i = 0
            candidate_index = 0

            for symbol in xmin_sorted:

                if c == 0 and symbol['label'] == '10':
                    # First candidate

                    candidate = symbol.copy()
                    c = 1  # found 1 candidate
                    candidate_index = i  # stores the candidate index

                elif c == 1 and symbol['label'] == '10':
                    # Second candidate

                    diff = abs(candidate['w'] / symbol['w'])

                    line1x = range(candidate['xmin'], candidate['xmax']+1)
                    line2x = range(symbol['xmin'], symbol['xmax']+1)
                    len_line1x = len(line1x)
                    len_line2x = len(line2x)
                    x_set = set(line1x) if len_line1x < len_line2x else set(line2x)
                    x_intersection = x_set.intersection(line1x if len_line1x >= len_line2x else line2x)

                    if x_intersection and diff <= 2 and diff >= 0.5:
                        c = 0

                        abacate = self.image.copy()

                        ymin = min(symbol['ymin'], candidate['ymin'])
                        ymax = max(symbol['ymax'], candidate['ymax'])

                        xmin = min(symbol['xmin'], candidate['xmin'])
                        xmax = max(symbol['xmax'], candidate['xmax'])

                        cropped = abacate[
                            ymin:ymax + 1,
                            xmin:xmax + 1
                        ]

                        pre = preprocessing.ImagePreprocessing()
                        resized = pre.resize(cropped)
                        resized = pre.binarization(resized)
                        resized = pre._255_to_1(resized)

                        symbol['image'] = resized
                        symbol['label'] = '30'
                        symbol['xmin'] = xmin
                        symbol['xmax'] = xmax
                        symbol['ymin'] = ymin
                        symbol['ymax'] = ymax
                        symbol['w'] = xmax - xmin
                        symbol['h'] = ymax - ymin
                        symbol['centroid'] = [xmin + (xmax-xmin) / 2, ymin + (ymax-ymin) / 2]

                        del xmin_sorted[candidate_index]

                    else:
                        candidate = symbol.copy()
                        candidate_index = i
                i += 1

        except BaseException as e:
            print(e)

        return xmin_sorted



import numpy as np
from mathreader import helpers
from mathreader.hme_parser.grammar import lex as lex
from mathreader.hme_parser.grammar import yacc as yacc

helpers_labels = helpers.get_labels()
labels = helpers_labels['labels_parser']


class CorrectGrammar():

    def __init__(self):
        pass

    def correct_grammar_lex(self, errors, latex, latex_list, index=0,
                            previous_errors=[]):
        """Fix grammar and lex errors.
        It tries to fix one error at the time.
        The error index is passed as parameter,
        otherwise, the first error (index=0) will be used.

        Args:
            errors (list): Errors found in previous step:
                either
                check_grammar_lex.py or check_grammar_sintax.py
            latex (list): First latex structure.
            latex_list (list): [description]
            index (int, optional): Index of error. Default to 0.
            previous_errors (list, optional):
                This list stores all previous symbol (index) attemptions
                to fix the error. Defaults to [].

        Returns:
            latex_string, errors, index: Updated data.
        """
        latex_string = ""

        helpers.debug("\n[base_grammar.py] correct_grammar_lex() | \
            List of errors: {0}".format(errors))

        previous_attemptions = []

        # It adds previous attempts to the array
        for error in previous_errors:
            helpers.debug('Previous error: {0}'.format(error))
            if error['attempts'] and len(error['attempts']) > 0:
                previous_attemptions.extend(error['attempts'])

        helpers.debug('Previous attemptions: {0}'.format(previous_attemptions))

        if len(errors) > 0:

            pos = errors[index]['pos']
            pos_list = errors[index]['pos_list']
            pred = errors[index]['prediction'].copy()

            subst = helpers.subst

            # When there's predictions it is a numpy array, not a list.
            if not isinstance(pred, list):

                json_label = 'labels_parser'

                def get_new_index(pred):
                    helpers.debug("[base_grammar.py] correct_grammar_lex() | \
                        Reset prediction of current symbol")
                    new_pred = pred.copy()
                    new_pred[0][np.argmax(pred)] = 0
                    helpers.debug("[base_grammar.py] correct_grammar_lex() | \
                        Gets new index and prediction from \
                        next index with higher prediction")
                    new_index = np.argmax(new_pred)
                    return new_index, new_pred

                def recur_get_new_index(pred):
                    new_index, pred = get_new_index(pred)
                    label_recog = helpers_labels[json_label][str(new_index)]
                    new_label = helpers_labels["labels_recognition"][label_recog]
                    new_identification = labels[new_label]

                    if new_identification in errors[index]['attempts'] or \
                    new_identification in previous_attemptions:

                        helpers.debug("[base_grammar.py] \
                            correct_grammar_lex() | \
                            New index is in previous attempts. Getting next.")

                        return recur_get_new_index(pred)

                    else:

                        if new_identification == '{':
                            new_identification = '\\{'
                        if new_identification == '}':
                            new_identification = '\\}'

                        return new_index, pred, new_identification

                new_index, new_pred, new_identification = recur_get_new_index(pred)
                helpers.debug("[base_grammar.py] correct_grammar_lex() | \
                    New symbol identification: %s " % new_identification)

                errors[index]['prediction'] = new_pred
                errors[index]['attempts'].append(new_identification)

                # Make it more 'Pythonic' later
                if new_identification in subst:

                    # list of substitutions
                    substitution_list = subst[new_identification]
                    for substitution_index in range(0, len(substitution_list)):
                        for substitution in substitution_list[substitution_index]:
                            if new_identification == substitution:
                                new_identification = substitution_list[substitution_index][substitution]

                '''
                    Updated 'latex' and 'latex_list'.
                    It is NOT a copy so, it is already updated.
                '''
                latex_list[pos_list] = new_identification
                latex[pos_list]['label'] = new_identification
                latex[pos_list]['prediction'] = new_pred
                latex_string = latex_string.join(latex_list)

        return {
            'latex_string': latex_string or "".join(latex_list),
            'errors': errors,
            'index': index
        }



from mathreader.hme_parser.grammar import lex
from mathreader.hme_parser.grammar.lex import tokens
import ply.yacc as yacc

def LatexParse(latex):
    error_list = []
    acceptance = []

    def p_gl(p):
        'GL : EL'
        # print('Aceita')
        acceptance.append('p_gl')

    def p_el_er_eq_er(p):
        'EL : ER EQ ER'
        acceptance.append('p_el_er_eq_er')

    def p_el_er_neq_er(p):
        'EL : ER NEQ ER'
        acceptance.append('p_el_er_neq_er')

    def p_el_er_eq(p):
        'EL : ER EQ'
        acceptance.append('p_el_er_eq')

    def p_el_er_neq(p):
        'EL : ER NEQ'
        acceptance.append('p_el_er_neq')

    def p_el_er(q):
        'EL : ER'
        acceptance.append('p_el_er')

    def p_er_er_plus_t(p):
        'ER : T PLUS ER'
        acceptance.append('p_er_er_plus_t')

    def p_er_er_min_t(p):
        'ER : T MIN ER'
        acceptance.append('p_er_er_min_t')

    def p_er_t(p):
        'ER : T'
        acceptance.append('p_er_t')

    def p_er_t_subsc_cdot_t(p):
        'T : SUBSC CDOT T'
        acceptance.append('p_er_t_subsc_cdot_t')

    def p_er_t_subsc(p):
        'T : SUBSC '
        acceptance.append('p_er_t_subsc')

    def p_er_subsc_pw_sub_subsc(p):
        'SUBSC : PW SUB SUBSC'
        acceptance.append('p_er_subsc_pw_sub_subsc')

    def p_er_subsc_pw(p):
        'SUBSC : PW'
        acceptance.append('p_er_subsc_pw')

    # def p_t_t_cdot_pw(p):
    #     'T : PW CDOT T'
    #     acceptance.append('p_t_t_cdot_pw')

    # def p_t_pw(p):
    #     'T : PW'
    #     acceptance.append('p_t_pw')

    def p_pw_pw_pow_fr(p):
        'PW : FR POW PW'
        acceptance.append('p_pw_pw_pow_fr')

    def p_pw_fr(p):
        'PW : FR'
        acceptance.append('p_pw_fr')

    def p_fr_lfrac(p):
        'FR : LFRAC OPENC ER CLOSEDC OPENC ER CLOSEDC'
        acceptance.append('p_fr_lfrac')

    def p_fr_s_lfrac(p):
        'FR : S LFRAC OPENC ER CLOSEDC OPENC ER CLOSEDC'
        acceptance.append('p_fr_s_lfrac')

    def p_fr_s(p):
        'FR : S'
        acceptance.append('p_fr_s')

    def p_s_sqrt(p):
        'S : SQRT OPENC ER CLOSEDC'
        acceptance.append('p_s_sqrt')

    def p_s_s_sqrt(p):
        'S : S SQRT OPENC ER CLOSEDC'
        acceptance.append('p_s_s_sqrt')

    def p_s_f_s (p):
        'S : S F'
        acceptance.append('p_s_f')

    def p_s_plus_f(p):
        'S : PLUS F'
        acceptance.append('p_s_lus_f')

    def p_s_min_f(p):
        'S : MIN F'
        acceptance.append('p_s_min_f')

    def p_s_f(p):
        'S : F'
        acceptance.append('p_s_f')

    def p_f_number(p):
        'F : NUMBER'
        acceptance.append('p_f_number')

    def p_f_var(p):
        'F : VAR'
        acceptance.append('p_f_var')

    def p_f_p(p):
        'F : P'
        acceptance.append('p_f_p')

    def p_p_openp(p):
        'P : OPENP ER CLOSEDP'
        acceptance.append('p_p_openp')

    def p_p_c(p):
        'P : C'
        acceptance.append('p_p_c')

    def p_c_openco(p):
        'C : OPENCO ER CLOSEDCO'
        acceptance.append('p_c_openco')

    def p_c_b(p):
        'C : B'
        acceptance.append('p_c_openc')

    def p_b_openc(p):
        'B : OPENC ER CLOSEDC'
        acceptance.append('p_b_openc')

    def p_error(p):
        global parser

        print("Syntax error in input!", p)

        erro = {
            'lexpos': None,
            'lineno': None,
            'type': None,
            'value': None
        }

        if p:
            erro.update({
                'lexpos': p.lexpos,
                'lineno': p.lineno,
                'type': p.type,
                'value': p.value
            })

        error_list.append(erro)

    parser = yacc.yacc()
    result = parser.parse(latex)

    if result:
        print("[yacc.py] Result: ", result)

    return error_list

if __name__ == "__main__":
    LatexParse(input(':'))


import ply.lex as lex

errors = []

tokens = (
'VAR',
'PLUS',
'MIN',
'OPENC',
'CLOSEDC',
'OPENP',
'CLOSEDP',
'OPENCO',
'CLOSEDCO',
'EQ',
'NEQ',
'CDOT',
'LFRAC',
'SQRT',
'NUMBER',
'POW',
'SUB'
)

t_VAR = r'(a|b|c|x|y|z|m|n)'
t_PLUS = r'\+'
t_MIN   = r'\-'
t_OPENC   = r'(\{|\\{)'
t_CLOSEDC  = r'(\}|\\})'
t_OPENP  = r'\('
t_CLOSEDP  = r'\)'
t_OPENCO = r'\['
t_CLOSEDCO = r'\]'
t_EQ = r'\='
t_NEQ = r'\\neq'
t_CDOT = r'\\cdot'
t_LFRAC = r'\\frac'
t_SQRT = r'\\sqrt'
t_NUMBER = r'[0-9]+(?:\.[0-9]+)?'
# t_NUMBER = r'^([0-9]+(?:\.[0-9]+)?)$'
t_POW = r'\^'
t_SUB = '_'

t_ignore = ' '

def t_error(t):
    errors.append((t.value[0], t.lexpos))
    t.lexer.skip(1)
    print('[lex.py] LEX ERROR: ', t)

# EOF handling rule
def t_eof(t):
    return None

lexer = lex.lex()

def LatexLexer(tstring):
    print('[lex.py] LEX STRING: ', tstring)
    """ Check lex errors

        Args:
        latex_string (str): Latex string.
        latex (list): First latex structure.
        latex_list (list): [description]

        Returns:
            errors (list): List of errors containg a tuple with (value, position) eg.: [('+',1)]
    """
    errors.clear()
    lexer.input(tstring)
    while True:
        tok = lexer.token()
        if not tok:
            return errors
    return errors

if __name__ == "__main__":
    LatexLexer(input(':'))


from mathreader.config import Configuration
import numpy as np
import cv2
import os

config = Configuration()

image_path = '/treatment/treated_data_three/not-number_testing_images.npz'
number = np.load(config.package_path + image_path)
number = number['arr_0']

label_path = '/treatment/treated_data_three/not-number_testing_labels.npz'
label_t = np.load(config.package_path + label_path)
label_t = label_t['arr_0']


def show_image(image):
    for i in range(0, len(image), 100):
        cv2.imshow("Image", image[i])
        print(label_t[i])
        cv2.waitKey(0)


show_image(number)

cv2.destroyAllWindows()




from mathreader.helpers import data_structures as DS
from mathreader import helpers
import re
import json
import numpy as np

helpers_labels = helpers.get_labels()
labels = helpers_labels['labels_parser']


class StructuralAnalysis:

    def __init__(self, symbol_list):
        self.symbols = symbol_list

    def analyze(self):

        symbols = self.symbols
        symbols = self.__preprocessing(symbols)
        tree = self.__main_parsing(symbols)
        if not tree:
            return
        tlist = self.__tree_to_list(tree)
        latex = self.__list_to_latex_obj(tlist)

        return {
            'latex': latex,
            'tree': tree,
            'tlist': tlist
        }

    def __preprocessing(self, symbols):
        helpers.debug('[parser.py] preprocessing()')
        xmin_sorted = sorted(symbols, key=lambda i: i['xmin'])
        symbols = xmin_sorted

        for i in range(0, len(symbols)):
            s = symbols[i]
            s['centroid'] = list(s['centroid'])
            s['checked'] = False

            if s['label'] in ['11', '13', '15']:
                s['type'] = 'Open'
            elif s['label'] in ['12', '14', '16']:
                s['type'] = 'Close'
            else:
                s['type'] = 'Normal'

            '''
                Centroid was too below or too above in these cases
                It was changed to: 1/3, 2/3
            '''

            '''
            [0-9] b
            s['ymin'] + (2/3) * (s['ymax'] - s['ymin'])

            y sqrt ()[]{}
            s['centroid'][1] = s['ymin'] + (1/3) * s['h']

            others
            s['centroid'][1] = s['ymin'] + ((s['ymax'] - s['ymin'])/2)
            '''

            '''
            Until validation:
                [0-9], b
            After validation:
                (), {}, [], sqrt
            '''
            if re.search("^[0-9]$", str(s['label'])) or \
                    s['label'] == '19' or s['label'] == '23' or \
                    s['type'] == 'Open' or s['type'] == 'Close':

                s['centroid_class'] = 'Ascending'
                s['centroid'][1] = s['ymin'] + (2/3) * (s['h'])  # 3/5

            elif s['label'] == '25':
                '''
                Until validation:
                    y sqrt ( { [
                After validation:
                    y sqrt
                '''
                s['centroid_class'] = 'Descending'
                s['centroid'][1] = s['ymin'] + (1/3) * s['h']

            else:
                s['centroid_class'] = 'Centred'
                s['centroid'][1] = s['ymin'] + ((s['h']) / 2)

            s['wall'] = {}
            s['wall']['top'] = -1
            s['wall']['bottom'] = 9999999999999
            s['wall']['left'] = -1
            s['wall']['right'] = 9999999999999

        return symbols

    def __main_parsing(self, symbols):
        helpers.debug('\n[parser.py] __main_parsing()')
        listin = symbols
        T = DS.Tree()
        Q = DS.Queue()
        S = DS.Stack()

        temp1 = 0
        temp2 = 0

        R = [[0, 0], [9999999999, 9999999999]]
        sstart = self.__sp(listin, R)

        if sstart == -1:
            return

        helpers.debug('\n[parser.py] __main_parsing() | \
            STARTING symbol index: %d ' % sstart)
        helpers.debug('[parser.py] __main_parsing() | \
            STARTING symbol label: %s ' % listin[sstart]['label'])

        s = listin[sstart]

        Q.enqueue(sstart)
        Q.enqueue(T.root_node)
        listin[sstart]['checked'] = True

        while not Q.is_empty():
            '''
            abc^{2-1}
            =============     ->    | EOBL  |
                        a            | c  |
            =============           | b |
                                    | a |
            '''

            '''
            abc^{2-1]
            =============     ->    |   |
                        2           |   |
            =============           | - |
                                    | 2 |
            '''

            helpers.debug('\n[parser.py] __main_parsing() \
                | find main baseline')

            while not Q.is_empty():
                temp1 = Q.dequeue()  # a, 2
                ParentNode = Q.dequeue()
                SymbolNode = DS.SymbolNode(listin[temp1])
                T.insert(SymbolNode, ParentNode, 'Node')
                S.push(temp1)  # a, 2
                S.push(SymbolNode)

                print('\n[parser.py] __main_parsing() | \
                    find baseline of symbol: ', temp1, listin[temp1]['label'])

                helpers.debug('\n[parser.py] __main_parsing() | \
                    temp2 hor...')
                temp2 = self.__hor(listin, temp1)  # b, -
                print('\n[parser.py] __main_parsing() | \
                    temp2: ', temp2, listin[temp2]['label'])

                while temp2 != -1:
                    print('[parser.py] __main_parsing() | \
                        ... while temp2')
                    listin[temp2]['checked'] = True
                    print('[parser.py] __main_parsing() | \
                        ... wall attributes of temp1: ', listin[temp1]['wall'])
                    listin[temp2]['wall'] = listin[temp1]['wall'].copy()
                    '''
                    a.wall = -1 -1 9999 9999

                    b.checked = true
                    b.wall = a.wall (-1 -1 9999 9999)

                    c.checked = true
                    c.wall = b.wall
                    ---------------------------------------------
                    -.checked = true
                    -.wall = 2.wall (wall da região super?)

                    '''

                    print('[parser.py] __main_parsing() | \
                        ...wall attributes of temp2: ', listin[temp2]['wall'])

                    SymbolNode = DS.SymbolNode(listin[temp2])
                    T.insert(SymbolNode, ParentNode, 'Node')
                    S.push(temp2)
                    S.push(SymbolNode)
                    listin[temp1]['wall']['right'] = listin[temp2]['xmin']

                    '''
                    a.wall.right = b.xmin
                    b.wall.right = c.xmin
                    '''

                    print('[parser.py] __main_parsing() | \
                        ...updated wall attributes \
                        of temp1: ', listin[temp1]['wall'])

                    temp1 = temp2  # b
                    temp2 = self.__hor(listin, temp1)  # c - 1
                    print('[parser.py] __main_parsing() | \
                        new temp2: ', temp2)

            S.push("EOBL")

            helpers.debug('\n[parser.py] __main_parsing() \
                | find secondary baseline')

            '''
            abc^2
            =============     ->    | EOBL  |
                        2           | c  |
            =============           | b |
                                    | a |
            '''
            while not S.is_empty():
                if S.peek() == "EOBL":
                    S.pop()
                SymbolNode = S.pop()
                temp1 = S.pop()  # c

                helpers.debug('[parser.py] __main_parsing() \
                    | symbol: %s ' % temp1)

                label = int(listin[temp1]['label'])
                helpers.debug('[parser.py] __main_parsing() \
                    | temp1 label: %s' % label)

                # 1/6
                upperThreshold = listin[temp1]['ymin'] + \
                    ((1/6.5) * listin[temp1]['h'])
                # 5/6
                lowerThreshold = listin[temp1]['ymin'] + \
                    ((5.5/6.5) * listin[temp1]['h'])

                '''
                    Changes in xmin and xmax because of the 'a'
                    When it overlaps the fraction
                '''
                leftThreshold = (
                        listin[temp1]['xmin'] + ((1/6) * listin[temp1]['w'])
                    ) \
                    if label != 10 else listin[temp1]['xmin']

                rightThreshold = (
                    listin[temp1]['xmax'] - ((1/6) * listin[temp1]['w'])
                    ) \
                    if label != 10 else listin[temp1]['xmax']

                R = [
                    {'above': [
                        [leftThreshold, listin[temp1]['wall']['top']],
                        [rightThreshold, upperThreshold]
                    ]},
                    {'below': [
                        [leftThreshold, lowerThreshold],
                        [rightThreshold, listin[temp1]['wall']['bottom']]
                    ]}
                ]

                for region in R:

                    # For each region, it looks for the initial symbol
                    reg = region[list(region.keys())[0]]
                    region_name = list(region.keys())[0]
                    helpers.debug('\n[parser.py] __main_parsing() | \
                        região: %s' % region_name)

                    # ( ) [ ] { } . * = neq + sqrt
                    operators = bool(label in range(11, 17) or
                    label in range(27, 31) or label == 17 or label == 23)

                    if (region_name == 'above' and not operators) or \
                    (region_name == 'below' and not operators):

                        temp2 = self.__start(listin, reg)

                        if temp2 != -1:
                            if not listin[temp2]['checked']:
                                listin[temp2]['checked'] = True
                                listin[temp2]['wall']['left'] = reg[0][0]
                                listin[temp2]['wall']['right'] = reg[1][0]
                                listin[temp2]['wall']['top'] = reg[0][1]
                                listin[temp2]['wall']['bottom'] = reg[1][1]

                                RelationNode = DS.RegionNode(list(region.keys())[0])
                                T.insert(RelationNode, SymbolNode, 'Node')
                                Q.enqueue(temp2)
                                Q.enqueue(RelationNode)

                '''
                    Changes in xmin and xmax because of the 'a'
                    When it overlaps the fraction
                '''
                R = [
                    {'contains': [
                        # left, top
                        [listin[temp1]['xmin'], listin[temp1]['ymin']],
                        # right, bottom
                        [listin[temp1]['xmax'], listin[temp1]['ymax']]
                    ]},
                    {'super': [
                        # left, top
                        [rightThreshold, listin[temp1]['wall']['top']],
                        # right, bottom
                        [listin[temp1]['wall']['right'], upperThreshold]
                    ]},
                    {'subsc': [
                        # left, top
                        [rightThreshold, lowerThreshold],
                        # right, bottom
                        [
                            listin[temp1]['wall']['right'],
                            listin[temp1]['wall']['bottom']
                        ]
                    ]}
                ]

                for region in R:
                    # Para cada região, busca o símbolo inicial
                    reg = region[list(region.keys())[0]]
                    region_name = list(region.keys())[0]
                    helpers.debug('\n[parser.py] __main_parsing() | \
                        região: %s' % region_name)

                    # - ( ) [ ] { } . * = neq +
                    operators = bool(label == 10 or
                    label in range(27, 31) or label == 17)

                    if (region_name == 'super' and not operators) or \
                    (region_name == 'subsc' and not operators) or \
                    (region_name == 'contains' and \
                    int(listin[temp1]['label']) == 23):

                        temp2 = self.__start(listin, reg)

                        if temp2 != -1:
                            if not listin[temp2]['checked']:
                                listin[temp2]['checked'] = True
                                listin[temp2]['wall']['left'] = reg[0][0]
                                listin[temp2]['wall']['right'] = reg[1][0]
                                listin[temp2]['wall']['top'] = reg[0][1]
                                listin[temp2]['wall']['bottom'] = reg[1][1]

                                RelationNode = DS.RegionNode(list(region.keys())[0])
                                T.insert(RelationNode, SymbolNode, 'Node')
                                Q.enqueue(temp2)
                                Q.enqueue(RelationNode)

        return T

    def __overlap(self, symbolIndex, top, bottom, listin):
        helpers.debug('\n\n[parser.py] __overlap()')
        listIndex = symbolIndex
        stop = False
        n = len(listin)

        helpers.debug('[parser.py] __overlap() | listIndex: %d ' % listIndex)

        if listin[symbolIndex]['label'] == '10':
            maxLength = listin[symbolIndex]['xmax'] - listin[symbolIndex]['xmin']
        else:
            maxLength = -1
        mainLine = -1

        helpers.debug('[parser.py] __overlap() | \
            mainLine: %d ' % mainLine)
        helpers.debug('[parser.py] __overlap() | \
            maxLength: %d ' % maxLength)

        while listIndex > 0 and stop == False:

            print('[parser.py] __overlap() | xmin, xmin',
                listin[listIndex-1]['xmin'],
                listin[symbolIndex]['xmin'])

            if listin[listIndex-1]['xmin'] <= listin[symbolIndex]['xmin']:
                listIndex = listIndex - 1  # stop = True
            else:
                stop = True  # listIndex = listIndex - 1

        helpers.debug('[parser.py] __overlap() | \
            listIndex: %d ' % listIndex)
        helpers.debug('[parser.py] __overlap() | \
            n: %d ' % n)
        helpers.debug('[parser.py] __overlap() | \
            top: %d ' % top)
        helpers.debug('[parser.py] __overlap() | \
            bottom: %d ' % bottom)

        line1x = range(listin[symbolIndex]['xmin'],
                        listin[symbolIndex]['xmax']+1)

        len_line1x = len(line1x)

        while listIndex < n and \
        listin[listIndex]['xmin'] < listin[symbolIndex]['xmax']:

            line2x = range(listin[listIndex]['xmin'],
                            listin[listIndex]['xmax']+1)
            len_line2x = len(line2x)
            x_set = set(line1x) if len_line1x < len_line2x else set(line2x)
            x_intersection = x_set.intersection(line1x if len_line1x >= len_line2x else line2x)
            min_line = min(len_line1x, len_line2x)

            print('\n[parser.py] __overlap() | ... \
                listIndex: ', listIndex)
            print('[parser.py] __overlap() | ... \
                label: ', listin[listIndex]['label'])
            print('[parser.py] __overlap() | ... \
                centroid: ', listin[listIndex]['centroid'])
            print('[parser.py] __overlap() | ... \
                xmin: ', listin[listIndex]['xmin'])
            print('[parser.py] __overlap() | ... \
                xmax: ', listin[listIndex]['xmax'])
            print('[parser.py] __overlap() | ... \
                max length: ', (
                    listin[listIndex]['xmax'] - listin[listIndex]['xmin']))
            print('[parser.py] __overlap() | ... \
                len(x_intersection): ', len(x_intersection))
            print('[parser.py] __overlap() | ... \
                min_line/2: ', min_line/2)

            if not listin[listIndex]['checked'] and \
                listin[listIndex]['label'] == '10' and \
                listin[listIndex]['centroid'][1] >= top and \
                listin[listIndex]['centroid'][1] <= bottom and \
                listin[listIndex]['xmin'] <= (listin[symbolIndex]['xmin'] + 8) and \
                len(x_intersection) > (min_line/2) and \
                (listin[listIndex]['xmax'] - listin[listIndex]['xmin']) > maxLength:
                    maxLength = (listin[listIndex]['xmax'] - listin[listIndex]['xmin'])
                    mainLine = listIndex

            listIndex += 1

        helpers.debug('[parser.py] __overlap() | listIndex: %d ' % listIndex)
        helpers.debug('[parser.py] __overlap() | mainLine: %d ' % mainLine)
        helpers.debug('[parser.py] __overlap() | maxLength: %d ' % maxLength)

        if mainLine == -1:
            return symbolIndex
        else:
            return mainLine

    def __start(self, listin, R):
        helpers.debug('\n[parser.py] __start()')
        print('[parser.py] __start() | region [R]: ', R)

        # Adicionei o round() | tirei o round
        left = R[0][0]
        top = R[0][1]
        right = R[1][0]
        bottom = R[1][1]

        helpers.debug('[parser.py] __start() | region [R]: \
            left: %d, right: %d,\
            top: %d, bottom: %d' %
            (left, right, top, bottom)
        )

        leftmostIndex = -1
        listIndex = 0
        overlapIndex = -1
        n = len(listin)

        while leftmostIndex == -1 and listIndex < n:
            helpers.debug('[parser.py] __start() | ... \
                symbol index: %d' % listIndex)
            helpers.debug('[parser.py] __start() | ... \
                symbol label: %s' % listin[listIndex]['label'])
            print('[parser.py]__start() | ... \
                symbol centroid: ', listin[listIndex]['centroid'])

            if not listin[listIndex]['checked'] and \
                listin[listIndex]['centroid'][0] >= left and \
                listin[listIndex]['centroid'][1] >= top and \
                listin[listIndex]['centroid'][0] <= right and \
                listin[listIndex]['centroid'][1] <= bottom:
                    leftmostIndex = listIndex
            else:
                listIndex = listIndex + 1

        helpers.debug('[parser.py] __start() | \
            leftmostIndex: %d' % leftmostIndex)

        if leftmostIndex == -1:
            return leftmostIndex
        else:
            return self.__overlap(leftmostIndex, top, bottom, listin)

    def __sp(self, listin, R):
        helpers.debug('\n[parser.py] __sp()')
        return self.__start(listin, R)

    def __hor(self, listin, index):
        print('\n[parser.py] __hor()')
        print('[parser.py] __hor() | symbol index: ',
            index)
        print('[parser.py] __hor() | symbol label: ',
            listin[index]['label'])

        global stop
        stop = False
        global a
        a = -1

        label = int(listin[index]['label'])

        right = listin[index]['wall']['right']

        # to avoid get symbols behind
        left = listin[index]['xmin']

        # to treat expoent and subscript
        # 1/6
        top = listin[index]['ymin'] + (listin[index]['h'] * (1/6.5))
        # 5/6
        bottom = listin[index]['ymin'] + (listin[index]['h'] * (5.5/6.5))

        # it doesn't have expoent and subscript
        if label == 10 or label in [27, 28, 29, 30]:
            top = listin[index]['wall']['top']
            bottom = listin[index]['wall']['bottom']

        # if it is square root, the left wall id xmax
        if label == 23:
            left = listin[index]['xmax']

        # if it is horizontal line or brackets
        if label in range(10, 17):
            R = [[listin[index]['xmax'], top], [right, bottom]]
            print('[parser.py] __hor() | R', R)
            a = self.__start(listin, R)
            stop = True

        else:

            helpers.debug('[parser.py] __hor() | top: %d, bottom: %d, \
                left: %d, right: %d' % (top, bottom, left, right))

            for s in range(0, len(listin)):

                checked = listin[s]['checked']

                if not checked:
                    symbol = listin[s]

                    helpers.debug('[parser.py] __hor() | ... \
                        symbol: %s' % symbol['label'])

                    helpers.debug('[parser.py] __hor() | ... \
                        symbol centroid: %s ' % symbol['centroid'])

                    helpers.debug('[parser.py] __hor() | ... \
                        symbol coordinates: xmin: %s xmax: %s \
                        ymin: %s ymax: %s' % (
                            symbol['xmin'], symbol['xmax'],
                            symbol['ymin'], symbol['ymax'])
                        )

                    if symbol['centroid'][0] >= left and \
                        symbol['centroid'][0] <= right and \
                        symbol['centroid'][1] <= bottom and \
                        symbol['centroid'][1] >= top:
                            helpers.debug('[parser.py] __hor() | \
                                ......... founded: %s' % s)
                            a = s
                            stop = True
                            break

                    helpers.debug('[parser.py] __hor() | a: %d ' % a)

        if a != -1:
            helpers.debug('[parser.py] __hor() | \
                a label: %s ' % listin[a]['label'])

        if stop and a != -1:
            helpers.debug('[parser.py] __hor() | ... before overlap')
            return self.__overlap(a, listin[a]['wall']['top'],
                                    listin[a]['wall']['bottom'], listin)
        else:
            return -1

    def __tree_to_list(self, tree, node=None):
        helpers.debug('[parser.py] __tree_to_list()')
        latex = []

        def recur(root_node):
            current = tree.root_node if not root_node else root_node

            if current is None:
                return

            if isinstance(current.data, str):
                latex.append(current.data)
            else:
                try:
                    real_label = labels[current.data['label']]

                    if real_label == '{':
                        real_label = '\\{'
                    if real_label == '}':
                        real_label = '\\}'

                    current.data['label'] = real_label
                    latex.append(current.data)
                except BaseException as e:
                    print('Exception: ', e)

            if current.node_type == 'RegionNode':
                latex.append('{')

            for node in current.children:
                recur(node)

            if current.node_type == 'RegionNode':
                latex.append('}')

        recur(node)

        if latex[0] == 'Expression':
            latex.remove('Expression')
            if latex[-1] == "}":
                latex.pop()
            if latex[0] == '{':
                latex.reverse()
                latex.pop()
                latex.reverse()

        return latex

    def __list_to_latex_obj(self, tlist):
        """ It turns the list into a dict:
            latex.append({
                'label': symbol[],
                'prediction': [],
                'type': ''
            })
        """

        helpers.debug('[parser.py] __list_to_latex_obj()')

        latex = []

        for symbol in tlist:
            if isinstance(symbol, dict):
                latex.append({
                    'label': symbol['label'],
                    'prediction': symbol['prediction'] \
                    if 'prediction' in symbol else [],
                    'type': symbol['type'] or ''
                })
            else:
                latex.append({
                    'label': symbol,
                    'prediction': [],
                    'type': 'context'
                })

        grammar = {
            '-': 'frac',
            'below': 'below',
            'sqrt': 'sqrt',
            'super': 'super',
            '*': 'mult',
            'subsc':  'subsc',
            'neq': 'neq'
        }

        subst = helpers.subst

        print('\n\n')
        latex = self.__token_substitution(latex, grammar, subst)
        print('\n\n')

        return latex

    def __token_substitution(self, latex, grammar, subst):

        def __list_substitution(i, nomatch, aux, initial_index, label,
                                substitution_list, substitution_index):

            for substitution in substitution_list[substitution_index]:
                helpers.debug('[parser.py] __list_to_latex_obj() | \
                    ......substitution: ')  # subsc
                helpers.debug('[parser.py] __list_to_latex_obj() | \
                    ......latex index: %d ' % i)  # subsc
                helpers.debug(substitution)  # subsc

                try:

                    helpers.debug('[parser.py] __list_to_latex_obj() | \
                        ......current latex: %s ' % latex[i]['label'])

                    # latex[i]['label'] = subst[subs][substitution]
                    if latex[i]['label'] == substitution:

                        helpers.debug('[parser.py] __list_to_latex_obj() | \
                            ......match ')

                        helpers.debug('[parser.py] __list_to_latex_obj() | \
                                    ......from: %s %s ' %
                                    (latex[i]['label'], substitution))

                        helpers.debug('[parser.py] __list_to_latex_obj() | \
                                    ......to: %s ' %
                                    substitution_list[substitution_index][substitution])

                        aux.append({
                            "index": i,
                            "label": substitution_list[substitution_index][substitution]
                        })

                        i += 1

                    else:

                        helpers.debug('[parser.py] __list_to_latex_obj() | \
                            ......no match ')

                        i -= 1
                        nomatch = True

                except IndexError as e:
                    helpers.debug('[parser.py] __list_to_latex_obj() | \
                        ......no match: IndexError ')
                    nomatch = True
                    break

                helpers.debug('[parser.py] __list_to_latex_obj() | \
                    ...... no match value: %s ' % nomatch)
                if nomatch:
                    i = initial_index
                    helpers.debug('[parser.py] __list_to_latex_obj() | \
                        ...... continue ......')
                    break
                helpers.debug('[parser.py] __list_to_latex_obj() | \
                    ...... next ......')
            return i, nomatch

        def __change_label(i, label, substitution_list):
            for substitution_index in range(0, len(substitution_list)):
                nomatch = False
                aux = []

                helpers.debug('[parser.py] __list_to_latex_obj() | \
                    ...substitutions: ')
                helpers.debug(substitution_list[substitution_index])

                initial_index = i

                i, nomatch = __list_substitution(i, nomatch, aux,
                                                initial_index,
                                                label, substitution_list,
                                                substitution_index)

                if not nomatch:
                    helpers.debug('[parser.py] __list_to_latex_obj() \
                        | match - updating value ')
                    for matched in aux:
                        latex[matched['index']]['label'] = matched['label']
            return i

        for i in range(0, len(latex)):
            if latex[i]['label'] in grammar:
                label = grammar[latex[i]['label']]

                if label in subst:
                    substitution_list = subst[label]  # list of substitutions
                    helpers.debug('[parser.py] __list_to_latex_obj() \
                        | substitution_list: ')
                    helpers.debug(substitution_list)

                    i = __change_label(i, label, substitution_list)

        return latex



import re
import json
import numpy as np
from mathreader.helpers import data_structures as DS
from mathreader.helpers.exceptions import GrammarError, LexicalError, SintaticError
from mathreader.hme_parser import check_grammar as cg
from mathreader.hme_parser import structural_analysis as sa
from mathreader import helpers


class Parser:
    def __init__(self, expression):
        self.symbols = expression['symbols']

    def __remove_contains(self, tstring):
        if isinstance(tstring, str):
            tstring = tstring.replace('contains', '')
        elif isinstance(tstring, list) and "contains" in tstring:
            tstring = tstring.remove('contains')
        return tstring

    def organize_latex_data(self, tstring):
        latex = []
        latex_list = []

        for symbol in tstring:
            if symbol['label'] != '' and symbol['label'] != 'contains':
                latex.append(symbol)
                latex_list.append(symbol['label'])

        latex_string = "".join(latex_list)
        latex_string = self.__remove_contains(latex_string)
        lstring = latex_string

        return {
            'latex': latex,
            'latex_list': latex_list,
            'latex_string': latex_string,
            'lstring': lstring
        }

    def to_parse(self):
        """
        Returns:
            {
                'latex': self.latex,
                'latex_list': self.latex_list,
                'latex_string': self.latex_string,
                'yacc_errors_history': self.yacc_errors_history,
                'lex_errors_history': self.lex_errors_history,
                'yacc_pure_errors': self.pure_yacc_errors,
                'lex_pure_errors': self.pure_lex_errors # It is not being added here
                'latex_before_cg': latex_before_cg,
                'tree': ...,
                'tlist': ...
            }
        """

        helpers.debug('[parser.py] to_parse()')

        try:
            structural_analysis = sa.StructuralAnalysis(self.symbols)
            structured_data = structural_analysis.analyze()

            if not structured_data:
                return

            latex_data = self.organize_latex_data(structured_data['latex'])

            check_grammar = cg.CheckGrammar()
            check_grammar_data = check_grammar.check(latex_data)

            data = {}
            data.update(check_grammar_data)
            data.update({'latex_before_cg': structured_data['latex']})
            data.update({
                'tree': structured_data['tree'],
                'tlist': structured_data['tlist']
            })

            return data

        except Exception as e:
            print('error')
            raise e


# for debuggind
if __name__ == "__main__":

    obj1 = {
        'symbols': [
            {'index': 2, 'xmin': 36, 'xmax': 102, 'ymin': 83, 'ymax': 161, 'w': 66, 'h': 78, 'centroid': [69.0, 80.5], 'label': '2'},
            {'index': 3, 'xmin': 109, 'xmax': 148, 'ymin': 48, 'ymax': 96, 'w': 39, 'h': 48, 'centroid': [128.5, 48.0], 'label': '2'},
            {'index': 0, 'xmin': 205, 'xmax': 245, 'ymin': 100, 'ymax': 147, 'w': 40, 'h': 47, 'centroid': [225.0, 73.5], 'label': '18'},
            {'index': 1, 'xmin': 300, 'xmax': 344, 'ymin': 88, 'ymax': 158, 'w': 44, 'h': 70, 'centroid': [322.0, 79.0], 'label': '5'}
        ]
    }

    obj2 = {
        'symbols': [
            {'index': 1, 'xmin': 143, 'xmax': 588, 'ymin': 197, 'ymax': 216, 'w': 445, 'h': 19, 'centroid': [365.5, 108.0], 'label': '11'},
            {'index': 4, 'xmin': 174, 'xmax': 534, 'ymin': 32, 'ymax': 199, 'w': 360, 'h': 167, 'centroid': [354.0, 99.5], 'label': '25'},
            {'index': 2, 'xmin': 280, 'xmax': 364, 'ymin': 92, 'ymax': 173, 'w': 84, 'h': 81, 'centroid': [322.0, 86.5], 'label': '19'},
            {'index': 0, 'xmin': 302, 'xmax': 428, 'ymin': 254, 'ymax': 358, 'w': 126, 'h': 104, 'centroid': [365.0, 179.0], 'label': '27'},
            {'index': 3, 'xmin': 379, 'xmax': 477, 'ymin': 60, 'ymax': 166, 'w': 98, 'h': 106, 'centroid': [428.0, 83.0], 'label': '20'}
        ]
    }

    a = Parser(obj1)
    print(a.to_parse())



from mathreader.hme_parser.grammar import yacc as yacc
from mathreader.hme_parser.grammar import lex as lex
from mathreader.hme_parser import correct_grammar as correct_grammar
from mathreader import helpers
from mathreader.helpers.exceptions import GrammarError, LexicalError, SintaticError
import numpy as np


class CheckLex():

    def __init__(self):
        self.__first_error = True
        self.latex_string = ""
        self.latex = ""
        self.latex_list = ""
        self.attempts = 0
        self.lex_error_list = None
        self.index = 0
        self.lex_errors_history = []
        self.pure_lex_errors = []

    def __locate_lex_error(self, lex_error_list):
        helpers.debug("\n[check_grammar_lex.py] __locate_lex_error()")

        error_list = lex_error_list.copy()
        latex = self.latex.copy()

        lex_errors = []
        lex_errors_history = self.lex_errors_history.copy()

        helpers.debug("[check_grammar_lex.py] __locate_lex_error() | \
            List of lex errors {0}: ".format(error_list))

        for error in error_list:

            if isinstance(error, dict):
                error = (error['label'], error['pos'])

            if error[1] != -1:

                helpers.debug("[check_grammar_lex.py] __locate_lex_error() | \
                    Locating error in latex_string.")

                count = 0
                count_list = 0

                latex_error_pos = error[1]  # position of char at string
                latex_error_token = error[0]  # It gets the token of the error

                # Search token (in latex) with position returned in error
                for symbol in latex:

                    if symbol['label'] != latex_error_token:

                        '''
                            Increment counter if current position is
                            the position containing the error
                        '''
                        count += len(symbol['label'])
                        count_list += 1

                    elif count == latex_error_pos:

                        lex_errors.append({
                            'pos': latex_error_pos,
                            'pos_list': count_list,
                            'label': symbol['label'],
                            'prediction': symbol['prediction'],
                            # It adds itself as a attempt of solution
                            'attempts': [symbol['label']]
                        })

                        lex_errors_history.extend(lex_errors)

                        break

            else:
                helpers.debug("[check_grammar_lex.py] __locate_lex_error() | \
                    Cannot fix the error. \
                    Use an automata do fix it. \
                    Logging the error: {0}".format(error))
                continue

        return lex_errors, lex_errors_history

    def check_correct_lex(self):
        """Check and correct lex errors

        Args:
            latex_string (str): Latex string.
            latex (list): First latex structure.
            latex_list (list): [description]

        Returns:
            {
                'latex': self.latex,
                'latex_list': self.latex_list,
                'latex_string': self.latex_string,
                'lex_errors_history': self.lex_errors_history,
                'lex_pure_errors': self.pure_lex_errors
            }
        """

        helpers.debug("\n[check_grammar_lex.py] check_correct_lex() | \
            attempts: %s" % self.attempts)

        second_lex_error_list = None
        lex_errors = []

        if not self.lex_error_list and \
        self.__first_error and \
        self.attempts < 3 and \
        self.latex_string:

            helpers.debug("\n[check_grammar_lex.py] check_correct_lex() | \
                There's no previous error. Searching the first one.")

            lex_error_list = lex.LatexLexer(self.latex_string)

            if lex_error_list:

                self.pure_lex_errors.extend(lex_error_list)
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | \
                    pure_lex_errors: {0}".format(self.pure_lex_errors))

                lex_errors, lex_errors_history = self.__locate_lex_error(lex_error_list)

                self.lex_error_list = lex_errors
                self.lex_errors_history = lex_errors_history
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | \
                    lex_error_list: {0}".format(self.lex_error_list))
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | \
                    lex_errors_history: {0}".format(self.lex_errors_history))

                self.__first_error = False

                self.__attempt_to_fix_error(lex_errors)

        elif self.lex_error_list and \
        not self.__first_error and \
        self.attempts < 3 and \
        self.latex_string:

            helpers.debug("\n[check_grammar_lex.py] check_correct_lex() | \
                There's previous error. Searching for new errors.")

            second_lex_error_list = lex.LatexLexer(self.latex_string)

            if second_lex_error_list:
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | \
                    New errors found.")

                self.pure_lex_errors.extend(second_lex_error_list)
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | \
                    pure_lex_errors: {0}".format(self.pure_lex_errors))

                ''' HUM... CONFERIR.
                    If new error is EOF error
                    Remove the error from the list. Takes the next one.
                '''
                if second_lex_error_list[0][1] == -1:
                    second_lex_error_list.reverse()
                    second_lex_error_list.pop()
                    second_lex_error_list.reverse()

                lex_errors, lex_errors_history = self.__locate_lex_error(second_lex_error_list)

                self.lex_error_list = lex_errors
                self.lex_errors_history = lex_errors_history
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | \
                    lex_error_list: {0}".format(self.lex_error_list))
                helpers.debug("[check_grammar_lex.py] check_correct_lex() | \
                    lex_errors_history: {0}".format(self.lex_errors_history))

                self.__attempt_to_fix_error(lex_errors)

        elif (self.lex_error_list and self.attempts >= 3) or \
        not self.latex_string:
            raise LexicalError({
                'latex': self.latex,
                'latex_list': self.latex_list,
                'latex_string': self.latex_string,
                'error': self.lex_error_list,  # Current error
                'errors_history': self.lex_errors_history,
                'pure_errors': self.pure_lex_errors
            })

        return {
            'latex': self.latex,
            'latex_list': self.latex_list,
            'latex_string': self.latex_string,
            'errors_history': self.lex_errors_history,
            'pure_errors': self.pure_lex_errors
        }

    def __attempt_to_fix_error(self, lex_errors):
        helpers.debug("[check_grammar_lex.py] self.__attempt_to_fix_error() \
            | Tries to fix the error.")

        # It tries to solve the FIRST error and returns an updated list of errors
        bg = correct_grammar.CorrectGrammar()

        # lex_errors: current error, self.lex_errors_history: all errors
        corrected_data = bg.correct_grammar_lex(lex_errors, self.latex,
                                                self.latex_list, 0,
                                                self.lex_errors_history)

        update_latex_string = corrected_data['latex_string']
        # Updated error with attempt
        self.lex_error_list = corrected_data['errors']
        self.index = corrected_data['index']

        # If there are remaining errors
        if self.lex_error_list:
            self.lex_errors_history = self.lex_error_list.copy()

        helpers.debug("[check_grammar_lex.py] self.__attempt_to_fix_error() | \
            Updated lex error: {0}".format(self.lex_error_list))
        helpers.debug("[check_grammar_lex.py] self.__attempt_to_fix_error() | \
            Updated lex error history: {0}".format(self.lex_errors_history))

        # if update_latex_string:
        self.latex_string = update_latex_string
        self.attempts += 1
        return self.check_correct_lex()



from mathreader.classification import classification as classification
from mathreader.image_processing import preprocessing as preprocessing
from mathreader.image_processing import postprocessing as postprocessing
from mathreader import helpers
import os
import re
import json
import numpy as np
import json
import cv2 as cv

helpers_labels = helpers.get_labels()
labels = helpers_labels['labels_parser']


class Recognize:

    def __init__(self, image):
        self.prediction = []
        self.image = None

        try:
            if isinstance(image, str):
                nparr = np.fromstring(image, np.uint8)
                img = cv.imdecode(nparr, cv.IMREAD_COLOR)
            else:
                img = image.copy()
            self.image = img
        except Exception as e:
            print("[recognize.py] __init__ | Exception:")
            raise e

    def __recognize(self, img):
        img = img.reshape(1, 28, 28, 1)
        return classification.fit(img)

    def to_recognize(self):

        expression = {}

        helpers.debug("[recognize.py] to_recognize | \
            Showing the image of the expression...\n")

        helpers.debug("[recognize.py] to_recognize | \
            Starting image preprocessing...\n")
        p = preprocessing.ImagePreprocessing()
        segmentation, normalized_image = p.treatment(self.image)

        helpers.debug("[recognize.py] to_recognize | \
            Showing preprocessed image\n")

        helpers.debug("[recognize.py] to_recognize | \
            Image preprocessing finished.\n")

        helpers.debug("[recognize.py] to_recognize | \
            Starting symbol classification...\n")

        try:
            for s in segmentation:

                helpers.debug('... segmentation ...')
                helpers.debug('... recognize ...')

                reconhecer = self.__recognize(s['image'])
                reconhecer['label'] = str(reconhecer['label'])

                s['label'] = reconhecer['label']
                s['prediction'] = reconhecer['prediction']

                symbol_prediction = {
                    'identity': labels[s['label']]
                }

                symbol_prediction.update(reconhecer)
                self.prediction.append(symbol_prediction)

                helpers.debug('[recognize.py] to_recognize | \
                    Símbolo: %s ' % labels[s['label']])

            helpers.debug("\n[recognize.py] to_recognize | \
                << Symbol classification finished.\n")

        except Exception as e:
            print("[recognize.py] to_recognize | Exception:")
            raise e

        pos = postprocessing.ImagePostprocessing(normalized_image)
        new = pos.segment_equality(segmentation)

        # teste
        for n in new:
            print('label: ', n['label'])

        expression['symbols'] = new

        return (expression, normalized_image)



import json


class GrammarError(Exception):
    def __init__(self, data=None, valor="Could not solve the grammar error"):
        self.valor = valor
        self.data = data

    def __repr__(self):
        return {
            'message': self.valor,
            'data': self.data
        }

    def __str__(self):
        return self.valor


class SintaticError(Exception):
    def __init__(self, data=None, valor="Could not solve the sintatic error"):
        self.valor = valor
        self.data = data

    def __repr__(self):
        return {
            'message': self.valor,
            'data': self.data
        }

    def __str__(self):
        return self.valor


class LexicalError(Exception):
    def __init__(self, data=None, valor="Could not solve the lexical error"):
        self.valor = valor
        self.data = data

    def __repr__(self):
        return {
            'message': self.valor,
            'data': self.data
        }

    def __str__(self):
        return self.valor



from mathreader import helpers
from mathreader.hme_parser import parser as parser
from mathreader.recognize import *
from mathreader.config import Configuration as config
import numpy as np
import cv2
import base64


class HME_Recognizer:

    def __init__(self):
        self.reset()

    def reset(self):
        self.image = None
        self.parsed_expression = ""
        self.processed_image = None
        self.predictions = None
        self.configurations = None
        self.expression_after_recognition = {}
        self.expression_after_parser = []
        self.expression_after_grammar = ""
        self.parser_tree = None
        self.parser_list = None
        self.yacc_errors_history = None
        self.yacc_pure_errors = None
        self.lex_errors_history = None
        self.lex_pure_errors = None

    def __to_parse(self, expression):
        try:
            parse = parser.Parser(expression)
            parsed_data = parse.to_parse()

            if not parsed_data:
                return

            self.yacc_errors_history = parsed_data['yacc_errors_history']
            self.yacc_pure_errors = parsed_data['yacc_pure_errors']
            self.lex_errors_history = parsed_data['lex_errors_history']
            self.lex_pure_errors = parsed_data['lex_pure_errors']
            self.latex_string_original = parsed_data['latex_string_original']
            self.expression_after_parser = parsed_data['latex_before_cg']
            self.expression_after_grammar = parsed_data['latex']
            self.parser_tree = parsed_data['tree']
            self.parser_list = parsed_data['tlist']

            return parsed_data['latex_string']

        except Exception as e:
            print("[api.py] __to_parse | Exception:")
            raise e

    def load_image(self, image, data_type='base64'):
        try:
            if data_type == 'base64':
                im_bytes = base64.b64decode(image)
                im_arr = np.frombuffer(im_bytes, dtype=np.uint8)
                img = cv2.imdecode(im_arr, flags=cv2.IMREAD_COLOR)

            elif data_type == 'path':
                img = cv2.imread(image, 0)

            elif data_type == 'byte':
                im_arr = np.frombuffer(image, dtype=np.uint8)
                img = cv2.imdecode(im_arr, flags=cv2.IMREAD_COLOR)

            else:
                raise Exception('Wrong file format. Should be: \
                    base64, path, byte.')

            self.image = img

        except BaseException as e:
            self.image = None
            raise e

    def recognize(self):
        try:
            img = self.image
            print(type(img))
            if isinstance(img, np.ndarray) and len(img) > 0:
                hme = Recognize(img)

                expression, image = hme.to_recognize()

                self.expression_after_recognition = expression.copy()
                self.predictions = hme.prediction

                parsed_expression = self.__to_parse(expression)

                if not parsed_expression:
                    raise Exception("Error")

                self.parsed_expression = parsed_expression
                self.processed_image = image

                return parsed_expression, image
            else:
                print("[api.py] recognize | \
                    Exception: You must enter an image")
                raise Exception("You must enter an image.")
        except Exception as e:
            print("[api.py] recognize | Exception:", e)
            raise e

    def get_predictions(self):
        return self.predictions

    def get_labels(self):
        labels = helpers.get_labels()
        return labels

    def get_expression_before_parser(self):
        return self.expression_before_parser

    def get_expression_after_parser(self):
        return self.expression_after_parser

    def get_expression_before_grammar(self):
        return self.expression_before_grammar

    def get_expression_after_grammar(self):
        return self.expression_after_grammar

    def get_yacc_errors(self):
        return self.yacc_errors_history

    def get_yacc_pure_errors(self):
        return self.yacc_pure_errors

    def get_lex_errors(self):
        return self.lex_errors_history

    def get_lex_pure_errors(self):
        return self.lex_pure_errors

    def get_latex_string_original(self):
        return self.latex_string_original

    def configuration(self):
        return config



# -*- coding: utf-8 -*-
from mathreader.config import Configuration
from mathreader import helpers
from tensorflow.keras.models import load_model
import tensorflow as tf
import numpy as np
import os

config = Configuration()

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['GLOG_minloglevel'] = '3'
print('\n')


def fit(image):

    labels = helpers.get_labels()

    try:
        model = load_model(config.package_path + '/ann_models/model/model_11-07-2020_23-54-57.h5')

        prediction = model.predict(image)
        index = np.argmax(prediction)
        print(index)
        label_rec = labels["labels_parser"][str(index)]

        return {
            'label': labels["labels_recognition"][label_rec],
            'prediction': prediction,
            'type': 'not-number'
        }

    except Exception as e:
        raise e



from mathreader import helpers
import numpy as np
import cv2
import imutils
import os


class ImagePreprocessing:

    def __init__(self, configs={}):
        self.configs = {
            'black': False,
            'dilate': False,
            'dataset': False,
            'erode': False,
            'resize': 'smaller'
        }

        if configs:
            self.configs.update(configs)

    def print_bounding_box(self, image, coordinates):
        image_rect = image
        x, y, w, h = coordinates
        cv2.rectangle(image_rect, (x, y), (x + w, y + h), (100, 100, 100), 2)
        # helpers.show_image(image_rect)
        return image_rect

    def save_image_255(self, image, name):
        img = cv2.convertScaleAbs(image, alpha=(255.0))
        cv2.imwrite('%s.jpg' % name, img)

    def to_gray_denoise(self, image):
        helpers.debug('[preprocessing.py] to_gray_denoise()')
        img = image.copy()

        if img.ndim == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        img = cv2.fastNlMeansDenoising(img, None, 5, 9)
        img = np.array(img)
        return img

    def invert(self, image):
        helpers.debug('[preprocessing.py] invert()')
        img = image.copy()
        return 255-img

    def binarization(self, image):
        helpers.debug('[preprocessing.py] binarization()')
        img = image.copy()
        img = self.invert(img)
        # img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 11, 2)
        img = cv2.adaptiveThreshold(img, 255, cv2.ADAPTIVE_THRESH_MEAN_C,
                                    cv2.THRESH_BINARY, 9, 2)
        img = self.invert(img)
        return img

    def normalize(self, image):
        helpers.debug('[preprocessing.py] normalize()')
        img = image.copy()
        img = self.to_gray_denoise(image)

        if not self.configs['black']:
            img = self.invert(img)

        kernel = np.ones((2, 2), np.uint8)

        if 'dilate' in self.configs:
            if self.configs['dilate']:
                img = cv2.dilate(img, kernel, iterations=2)

        if 'erode' in self.configs:
            if self.configs['erode']:
                img = cv2.erode(img, kernel, iterations=1)

        return img

    def _255_to_1(self, image):
        helpers.debug('[preprocessing.py] _255_to_1()')
        img = image.copy()
        return (img / 255)

    def resize(self, image):
        helpers.debug('[preprocessing.py] resize()')
        old_size = image.shape[:2]  # (height, width)
        height, width = old_size[0], old_size[1]

        ratio = float(26) / max(old_size)
        size = tuple([int(x*ratio) for x in old_size])

        size_height, size_width = size[0], size[1]
        size_height = size_height if size_height > 0 else 1
        size_width = size_width if size_width > 0 else 1

        division_height = int(height/2)
        division_width = int(width/2)
        around_w = round(width * 20 / 100)
        around_h = round(height * 20 / 100)

        middle_width = [
            image[division_height][division_width-1],
            image[division_height][division_width],
            image[division_height][division_width+1]
        ]

        middle_height = []
        for a in range(division_height-around_h, division_height+around_h):
            middle_height.append(
                image[a][division_width],
            )

        middle_width = []
        for b in range(division_width-around_w, division_width+around_w):
            middle_width.append(
                image[division_height][division_width],
            )

        helpers.debug('[preprocessing.py] segment() | \
            before line and sqrt processing')
        if size_height <= 15 and size_width >= 20 and \
            (any(i > 0.0000 for i in middle_height) or
                any(i > 0.0000 for i in middle_width)):

            # For horizontal line
            nsize = 4 if size_height < 5 else size_height
            nsize = 10 if size_height > 10 else size_height
            print('hor ', size_height)
            new_size = tuple([int(nsize), 26])

        else:
            if size_width / size_height >= 2:
                # For rectangle (sqrt)
                kernel = np.ones((2, 2), np.uint8)
                image = cv2.dilate(image, kernel, iterations=7)
                # xinit = int(width * 2 / 100) # validation
                # xend = int(width * 65 / 100) # validation
                xinit = int(width * 5 / 100)
                xend = int(width * 85 / 100)
                image = image[0:height, xinit:xend]

            new_size = size

        helpers.debug('[preprocessing.py] segment() | \
            after line and sqrt processing')

        helpers.debug('[preprocessing.py] segment() | \
            before resize')

        if self.configs['resize'] == 'smaller':
            image = cv2.resize(image.copy(), (new_size[1], new_size[0]),
                               interpolation=cv2.INTER_AREA)
        elif self.configs['resize'] == 'bigger':
            image = cv2.resize(image.copy(), (new_size[1], new_size[0]),
                               interpolation=cv2.INTER_LINEAR)
        helpers.debug('[preprocessing.py] segment() | after resize')

        # Cria borda ao redor do símbolo e normaliza para 28x28 px
        helpers.debug('[preprocessing.py] segment() | before border')
        delta_w = 28 - new_size[1]
        delta_h = 28 - new_size[0]
        top, bottom = delta_h//2, delta_h-(delta_h//2)
        left, right = delta_w//2, delta_w-(delta_w//2)
        color = [0, 0, 0]
        image = cv2.copyMakeBorder(image.copy(), top, bottom,
                                   left, right, cv2.BORDER_CONSTANT,
                                   value=color)

        helpers.debug('[preprocessing.py] segment() | after border')
        return image

    def resize_full_image(self, image):
        helpers.debug('[preprocessing.py] resize_full_image()')
        h, w = image.shape[:2]
        if w > 4000:
            width = int(w * 20 / 100)
            r = width / float(w)
            size = (width, int(h * r))
            image = cv2.resize(image, size)
        return image

    def segment(self, img):
        helpers.debug('[preprocessing.py] segment()')
        image = img.copy()
        symbols = []
        cnts, somethingElse = cv2.findContours(image.copy(), cv2.RETR_EXTERNAL,
                                               cv2.CHAIN_APPROX_SIMPLE)

        helpers.debug('[preprocessing.py] segment() | contours founded:')
        helpers.debug(len(cnts))

        for i in range(len(cnts)):
            # It was 0 and was changed to 10 to try reducing the noise
            if(cv2.contourArea(cnts[i]) < 10):
                continue

            if(self.configs['dataset'] and len(cnts) > 1):
                continue

            try:
                # Draw contour in new image (mask)
                mask = np.zeros_like(image)
                cv2.drawContours(mask, cnts, i, (255, 255, 255), -50)
                out = np.zeros_like(image)

                '''
                At the position where the mask is 255
                it paints the normalized position
                with the same positions from mask were it is 255

                I.e. mask has the inner part of "2" painted,
                but the normalized doesn't have it.

                It was changed to > 0 instead of 255
                this prevents the image from having to be binarized
                '''
                out[mask > 0] = image[mask > 0]

                # # Get bounding box coordinates
                _x, _y, _w, _h = cv2.boundingRect(cnts[i])

                # For now, it's worthless
                # ALL points where the mask == 255
                # list_y, list_x = np.where(out > 0)
                # (topx, topy) = (np.min(list_x), np.min(list_y))
                # (bottomx, bottomy) = (np.max(list_x), np.max(list_y))

                # Crop the image
                ycrop = _y + _h + 1
                xcrop = _x + _w + 1
                cropped = out[_y: ycrop, _x: xcrop]

                resized = self.resize(cropped)

                # Test - It was not here during validation
                binarized = self.binarization(resized)
                result_image = self._255_to_1(binarized)

                # result_image = self._255_to_1(resized)
                helpers.show_image(result_image)

                attributes = {
                    'index': i,
                    'image': result_image.copy(),
                    'xmin': _x,
                    'xmax': _x+_w,
                    'ymin': _y,
                    'ymax': _y+_h,
                    'w': _w,
                    'h': _h,
                    'centroid': [(_x + (_x+_w))/2, (_y + (_y+_h))/2]
                }

                symbols.append(attributes)

                mask = None
                out = None
                cropped = None
                resized = None
                binarized = None
                result_image = None

                # self.image = self.print_bounding_box(image, (_x, _y, _w, _h))
                self.image = image

            except BaseException as e:
                print(e)
                continue

        return (symbols, self.image)

    def mnist(self, img):
        try:
            if type(img) is str:
                image = cv2.imread(img)
            else:
                image = img

            normalized = self.normalize(image)
            result_image = self._255_to_1(normalized)
            return result_image
        except BaseException as e:
            print(e)
            return None

    def treatment_without_segmentation(self, img):
        try:
            if type(img) is str:
                image = cv2.imread(img)
            else:
                image = img

            normalized = self.normalize(image)
            resized = self.resize(normalized)
            result_image = self._255_to_1(resized)
            return result_image
        except BaseException as e:
            print(e)
            return None

    def treatment(self, img):
        helpers.debug('[preprocessing.py] treatment()')
        try:
            if type(img) is str:
                image = cv2.imread(img)
            else:
                image = img

            image = self.resize_full_image(image)
            normalized = self.normalize(image)
            self.image = normalized.copy()

            symbols = self.segment(normalized)

            return symbols
        except BaseException as e:
            print(e)
            return []


if __name__ == "__main__":

    p = ImagePreprocessing()
    segmentation, image = p.treatment('images/numbers/teste10.png')

    try:
        for s in segmentation:
            s.__delitem__('image')
    except Exception as identifier:
        print(identifier)

    print(segmentation)



import setuptools

with open("longdescription.txt", "r") as f:
    long_description = f.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

try:
    setuptools.setup(
        name = 'mathreader',
        version = '0.162',
        author = 'Caroline Reis',
        author_email = 'caroline.reis@rede.ulbra.br',
        long_description = long_description,
        #long_description_content_type="text/markdown",
        url="https://github.com/carolreis/mathreader",
        packages = setuptools.find_packages(),
        install_requires=required,
        include_package_data=True,
        python_requires='>=3.6'
    )
except Exception as e:
    pass



from mathreader.config import Configuration
import os
import cv2 as cv
import idx2numpy
import json

config = Configuration()

subst = {
    'frac': [{
        '-': "\\frac",
        'above': '',
    }],
    'below': [{
        'below': ''
    }],
    'sqrt': [{
        'sqrt': '\\sqrt'
    }],
    'super': [{
        'super': '^'
    }],
    'mult': [{
        '*': '\cdot'
    }],
    'subsc': [
        {
            'subsc': '',
            '{': '',
            '.': '.',
            '}': ''
        },
        {
            'subsc': '_'
        }
    ],
    'neq': [{
        'neq': '\\neq'
    }]
}


def create_dir(name):
    dirname = name
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    return dirname


def show_image(image):
    conf = config.get_configs()
    if conf["application"]["debug_mode_image"] == "active":
        cv.namedWindow("Imagem", cv.WINDOW_NORMAL)
        # cv.resizeWindow("Imagem", 725, 325)
        # cv.moveWindow("Imagem", 40,525)
        cv.imshow("Imagem", image)
        cv.waitKey(0)
        cv.destroyAllWindows()


def save_image(self, dirname, roi, i):
    filename = "char" + str(i) + ".png"
    cv2.imwrite(dirname + "/" + filename, roi)


def get_labels():

    config_path = '/docs/config/config_all.json'

    try:
        with open(config.package_path + config_path) as json_file:
            labels_json = json_file.read()
            labels_dict = json.loads(labels_json)
            labels = labels_dict
    except Exception as e:
        print(e)
        labels = {}

    return labels


def save_json_file(path, data):
    with open(path, 'w') as f:
        f.write(json.dumps(data))


def debug(data):
    conf = config.get_configs()
    if conf and "application" in conf:
        if conf["application"]["debug_mode"] == "active":
            print(data)



from mathreader.hme_parser.grammar import yacc as yacc
from mathreader.hme_parser.grammar import lex as lex
from mathreader.hme_parser import correct_grammar as correct_grammar
from mathreader.hme_parser import check_grammar_lex as check_grammar_lex
from mathreader.helpers.exceptions import GrammarError, LexicalError, SintaticError
from mathreader import helpers
import numpy as np


class CheckSintax():

    def __init__(self):
        self.__first_error = True
        self.latex_string = ""
        self.latex = ""
        self.latex_list = ""
        self.attempts = 0
        self.index = 0

        self.lex_errors_history = []
        self.yacc_errors_history = []

        self.yacc_error_list = None
        self.lex_error_list = None

        self.pure_yacc_errors = []
        self.pure_lex_errors = []

    def set_lex_data(self, check_lex_data):
        self.latex = check_lex_data['latex']
        self.latex_list = check_lex_data['latex_list']
        self.latex_string = check_lex_data['latex_string']
        self.lex_errors_history = check_lex_data['errors_history']
        self.pure_lex_errors = check_lex_data['pure_errors']

    def __locate_grammar_error(self, yacc_error_list):
        helpers.debug("\n[check_grammar_sintax.py] __locate_grammar_error() | \
            Locating all errors and creating a data structure.")

        yacc_error_list = yacc_error_list.copy()
        latex = self.latex.copy()

        yacc_errors = []
        yacc_errors_history = self.yacc_errors_history.copy()

        helpers.debug("[check_grammar_sintax.py] __locate_grammar_error() | \
            Errors: {0}".format(yacc_error_list))

        for error in yacc_error_list:

            if error['value'] is not None:

                helpers.debug("[check_grammar_sintax.py] \
                    __locate_grammar_error() | ...for() value not none")

                count = 0
                count_list = 0

                latex_error_pos = error['lexpos']
                latex_error_token = error['value']

                for symbol in latex:

                    if symbol['label'] == latex_error_token and \
                    count == latex_error_pos:

                        yacc_errors.append({
                            'pos': latex_error_pos,
                            'pos_list': count_list,
                            'label': symbol['label'],
                            'prediction': symbol['prediction'],
                            # It adds itself as a attempt of solution
                            'attempts': [symbol['label']]
                        })

                        yacc_errors_history.extend(yacc_errors)

                        break

                    count += len(symbol['label'])
                    count_list += 1

            else:
                helpers.debug("Use automata to fix")
                continue

        return yacc_errors, yacc_errors_history

    def check_correct_grammar(self):
        """Check and correct lex errors

        Args:
            latex_string (str): Latex string.
            latex (list): First latex structure.
            latex_list (list): [description]

        Returns:
            {
                'latex': self.latex,
                'latex_list': self.latex_list,
                'latex_string': self.latex_string,
                'yacc_errors_history': self.yacc_errors_history,
                'lex_errors_history': self.lex_errors_history,
                'yacc_pure_errors': self.pure_yacc_errors,
                'lex_pure_errors': self.pure_lex_errors # Não está sendo adicionado aqui
            }
        """

        helpers.debug("\n----------------------------------------------------")
        helpers.debug("[check_grammar_sintax.py] check_correct_grammar() | \
            attempts: %s" % self.attempts)

        second_yacc_error_list = None
        yacc_errors = []

        if not self.yacc_error_list and \
        self.__first_error and \
        self.attempts < 3 and \
        self.latex_string:

            helpers.debug("[check_grammar_sintax.py] check_correct_grammar() |\
                There's no previous errors")

            yacc_error_list = yacc.LatexParse(self.latex_string)

            if yacc_error_list:

                self.pure_yacc_errors.extend(yacc_error_list)
                helpers.debug("[check_grammar_sintax.py] \
                    check_correct_grammar() | \
                    pure_yacc_errors: {0}".format(self.pure_yacc_errors))

                # It treats the EOF error
                if yacc_error_list[0]['lexpos'] is None:

                    yacc_error_list[0].update({
                        'lexpos': len(self.latex_string) - 1
                    })

                    yacc_error_list[0].update({
                        'value': self.latex_string[-1]
                    })

                yacc_errors, yacc_errors_history = self.__locate_grammar_error(yacc_error_list)

                self.yacc_error_list = yacc_errors
                self.yacc_errors_history = yacc_errors_history
                helpers.debug("[check_grammar_yacc.py] check_correct_yacc() | \
                    yacc_error_list: {0}".format(self.yacc_error_list))
                helpers.debug("[check_grammar_yacc.py] check_correct_yacc() | \
                    yacc_errors_history: {0}".format(self.yacc_errors_history))

                self.__first_error = False

                self.__attempt_to_fix_error(yacc_errors)

        elif self.yacc_error_list and \
        not self.__first_error and \
        self.attempts < 3 and \
        self.latex_string:

            '''
                After correcting the first grammar error,
                it's necessary to check wheter the correction
                generated lexical errors or not
            '''

            helpers.debug("\n[check_grammar_sintax.py] \
                check_correct_grammar() | \
                There were errors before")

            self.__find_lexical_errors()
            # Execution gets here when there's no more lexical errors

            second_yacc_error_list = yacc.LatexParse(self.latex_string)

            if second_yacc_error_list:
                helpers.debug("\n[check_grammar_sintax.py] \
                    check_correct_grammar() | \
                    New errors found")

                self.pure_yacc_errors.extend(second_yacc_error_list)
                helpers.debug("\n[check_grammar_sintax.py] \
                    check_correct_grammar() | \
                    pure_yacc_errors: {0}".format(self.pure_yacc_errors))

                # It handles the EOF error
                if second_yacc_error_list[0]['lexpos'] is None:

                    second_yacc_error_list[0].update({
                        'lexpos': len(self.latex_string) - 1
                    })

                    second_yacc_error_list[0].update({
                        'value': self.latex_string[-1]
                    })

                yacc_errors, yacc_errors_history = self.__locate_grammar_error(second_yacc_error_list)

                self.yacc_error_list = yacc_errors
                self.yacc_errors_history = yacc_errors_history
                helpers.debug("[check_grammar_yacc.py] check_correct_yacc() | \
                    yacc_error_list: {0}".format(self.yacc_error_list))
                helpers.debug("[check_grammar_yacc.py] check_correct_yacc() | \
                    yacc_errors_history: {0}".format(self.yacc_errors_history))

                self.__attempt_to_fix_error(yacc_errors)

        elif (self.yacc_error_list and self.attempts >= 3) or \
        not self.latex_string:

            helpers.debug("\n[check_grammar_sintax.py] \
                check_correct_grammar() | GrammarError")

            raise GrammarError({
                'latex': self.latex,
                'latex_list': self.latex_list,
                'latex_string': self.latex_string,
                'error': self.yacc_error_list,  # Current error
                'errors_history': self.yacc_errors_history,
                'pure_errors': self.pure_yacc_errors
            })

        # It will be returned when the grammar is solved
        return {
            'latex': self.latex,
            'latex_list': self.latex_list,
            'latex_string': self.latex_string,
            'yacc_errors_history': self.yacc_errors_history,
            'lex_errors_history': self.lex_errors_history,
            'yacc_pure_errors': self.pure_yacc_errors,
            'lex_pure_errors': self.pure_lex_errors
        }

    def __find_lexical_errors(self):
        helpers.debug("\n................FIND LEXICAL ERRORS................")

        cgl = check_grammar_lex.CheckLex()

        cgl.latex_string = self.latex_string
        cgl.latex = self.latex
        cgl.latex_list = self.latex_list
        cgl.attempts = 0

        check_lex_data = cgl.check_correct_lex()
        # If the executtion got here is because the error was solved

        self.latex = check_lex_data['latex']
        self.latex_list = check_lex_data['latex_list']
        self.latex_string = check_lex_data['latex_string']
        self.pure_lex_errors = check_lex_data['pure_errors']
        lex_errors_history = check_lex_data['errors_history']
        self.lex_errors_history.extend(lex_errors_history)

        helpers.debug("...................................................")

    def __attempt_to_fix_error(self, yacc_errors):
        helpers.debug('[check_grammar_sintax.py] __attempt_to_fix_error()')

        '''
            It tries to solve the FIRST error and
            returns an updated list of errors
        '''
        bg = correct_grammar.CorrectGrammar()

        # It join Lex and Yacc's attempts at solutions.
        fix_attempts = self.lex_errors_history.copy()
        fix_attempts.extend(self.yacc_errors_history)

        corrected_data = bg.correct_grammar_lex(yacc_errors, self.latex,
                                                self.latex_list, 0,
                                                fix_attempts)

        updated_latex_string = corrected_data['latex_string']
        # Updated error with attempt
        self.yacc_error_list = corrected_data['errors']
        self.index = corrected_data['index']

        # It there's remaining errors
        if self.yacc_error_list:
            self.yacc_errors_history = self.yacc_error_list.copy()

        helpers.debug("[check_grammar_yacc.py] \
            self.__attempt_to_fix_error() | \
            Updated yacc error: {0}".format(self.yacc_error_list))
        helpers.debug("[check_grammar_yacc.py] \
            self.__attempt_to_fix_error() | \
            Updated yacc error history: {0}".format(self.yacc_errors_history))

        # if updated_latex_string:
        self.latex_string = updated_latex_string
        self.attempts += 1
        return self.check_correct_grammar()



import mathreader
import json


class Configuration:
    package_path = mathreader.__path__[0]
    config_path = package_path + '/docs/config/config.json'

    def get_configs(self):
        configs = {}
        try:
            with open(self.config_path) as json_file:
                config = json_file.read()
                if config:
                    configs = json.loads(config)
                else:
                    configs = {}
        except Exception as e:
            print("[config.py] Error while opening configurations: ", e)
        return configs

    def set_app_debug_mode(self, value):
        print("[config.py] set_app_debug_mode() ")
        if isinstance(value, str) and value != "":
            self.update_data('debug_mode', value)

    def set_app_debug_mode_image(self, value):
        print("[config.py] set_app_debug_mode_image() ")
        if isinstance(value, str) and value != "":
            self.update_data('debug_mode_image', value)

    def update_data(self, key, value):
        print("[config.py] update_data()")
        try:
            configs = self.get_configs()
            if 'application' in configs:
                configs['application'].update({key:  value})
            else:
                configs.update({'application': {
                    key: value
                }})
            with open(self.config_path, 'w') as f:
                f.write(json.dumps(configs))
            print('[config.py] update_data() | config updated: ', configs)
        except Exception as e:
            print("[config.py] Error while updating config json")
            raise e



labels = {
    0: '0',
    1: '1',
    2: '2',
    3: '3',
    4: '4',
    5: '5',
    6: '6',
    7: '7',
    8: '8',
    9: '9',
    10: '=',
    11: '-',
    12: '(',
    13: ')',
    14: '[',
    15: ']',
    16: '{',
    17: '}',
    18: '+',
    19: 'a',
    20: 'b',
    21: 'c',
    22: 'div',
    23: 'm',
    24: 'n',
    25: 'sqrt',
    26: 'times',
    27: 'x',
    28: 'y',
    29: 'z',
    30: '*'
}


class StackNode:
    def __init__(self, data=None):
        self.data = data
        self.next = None


class Stack:
    def __init__(self):
        self.top = None
        self.size = 0

    def push(self, data):
        node = StackNode(data)
        if self.top:
            node.next = self.top
            self.top = node
        else:
            self.top = node
        self.size += 1

    def pop(self):
        if self.top:
            data = self.top.data
            self.size -= 1
            if self.top.next:
                self.top = self.top.next
            else:
                self.top = None
            return data
        else:
            return None

    def peek(self):
        if self.top:
            return self.top.data
        else:
            return None

    def is_empty(self):
        return self.size == 0


class QueueNode:
    def __init__(self, data=None, _next=None, _prev=None):
        self.data = data
        self.next = _next
        self.prev = _prev


class Queue:
    def __init__(self):
        self.head = None
        self.tail = None
        self.count = 0

    def enqueue(self, data):
        new_node = QueueNode(data, None, None)
        if self.head is None:
            self.head = new_node
            self.tail = self.head
        else:
            new_node.prev = self.tail
            self.tail.next = new_node
            self.tail = new_node

        self.count += 1

    def dequeue(self):
        current = self.head
        if self.count == 1:
            self.count -= 1
            self.head = None
            self.tail = None
        elif self.count > 1:
            self.head = self.head.next
            self.head.prev = None
            self.count -= 1

        return current.data

    def peek(self):
        if self.head:
            return self.head.data
        else:
            return None

    def is_empty(self):
        return self.count == 0


class TreeNode:
    def __init__(self, data=None):
        self.data = data
        self.children = []
        self.node_type = 'TreeNode'

    def __str__(self):
        return str(self.data)


class RegionNode:
    def __init__(self, data=None):
        self.data = data
        self.children = []
        self.node_type = 'RegionNode'

    def __str__(self):
        return str(self.data)


class SymbolNode:
    def __init__(self, data=None):
        self.data = data
        self.children = []
        self.node_type = 'SymbolNode'

    def __str__(self):
        return str(self.data)


class Tree:

    def __init__(self):
        self.root_node = RegionNode('Expression')

    def preorder(self, root_node=None):
        current = self.root_node if not root_node else root_node
        if current is None:
            return

        if isinstance(current.data, str):
            print(current.data)
        else:
            try:
                print(labels[current.data['label']])
            except Exception as e:
                print('')

        for node in current.children:
            self.preorder(node)

    def insert(self, data, insert_node=None, _type='TreeNode'):

        if _type == 'Region':
            node = RegionNode(data)
        elif _type == 'Symbol':
            node = SymbolNode(data)
        elif _type == 'TreeNode':
            node = TreeNode(data)
        elif _type == 'Node':
            node = data
        current = insert_node if insert_node else self.root_node
        parent = current

        parent.children.append(node)

        return node


if __name__ == "__main__":
    tree = Tree()
    a = tree.insert('A')
    b = tree.insert('B')
    c = tree.insert('C')
    p = tree.insert('=P', c)
    d = tree.insert('D', b)
    eight = tree.insert('8', b)
    haha = tree.insert('haha', b)
    tree.preorder()
